'From Squeak3.8alpha of ''17 July 2004'' [latest update: #6313] on 9 October 2004 at 8:05:54 pm'!
"Change Set:		Compiler-ajh
Date:			14 Map 2004
Author:			Anthony Hannan <ajh18@cornell.edu>

This adds the closure compiler.  It uses SmaCC and the RB AST.  Closures-ajh must be filed in before this.  To use the closure compiler set Preferences compileBlocksAsClosures to true.  Read the Compiler class comment for more.

MD: removed all additions to AST classes, smacc etc...

"!


!Compiler commentStamp: 'ajh 3/26/2003 00:37' prior: 0!
The compiler accepts Smalltalk source code and compiles it with respect to a given class. The user of the compiler may supply a context so that temporary variables are accessible during compilation. If there is an error, a requestor (usually a kind of StringHolderController) is sent the message notify:at:in: so that the error message can be displayed. If there is no error, then the result of compilation is a MethodNode, which is the root of a parse tree whose nodes are kinds of ParseNodes. The parse tree can be sent messages to (1) generate code for a CompiledMethod (this is done for compiling methods or evaluating expressions); (2) pretty-print the code (for formatting); or (3) produce a map from object code back to source code (used by debugger program-counter selection).

----

Closure Compiler
by Anthony Hannan 3/13/2003

The Compiler class (like before) serves as the interface to the compiler.  If Preferences compileBlocksAsClosures is true it will run the new closure compiler.  The closure compiler classes reside in their own system category called 'Compiler', which is broken down into four minor categories: 'Compiler-Syntax', 'Compiler-Semantics', 'Compiler-IR', and 'Compiler-Bytecodes'.  The closure compiler also uses classes in 'SmaCC-Runtime' and the Refactory abstract syntax tree in 'Refactory-Parser'.

Like most compilers, the closure compiler translates source text into bytecodes in a sequence of phases or transformations.

Phase 1 - Scan and parse text into abstract syntax tree (AST)

SqueakScanner and SqueakParser in 'Compiler-Syntax' transform a method or do-it text into a Refactory abstract syntax tree.  SqueakScanner and SqueakParser classes were automatically generated using the SmaCC parser generator from a token and grammar specification derived from the StScanner and StParser specifications that comes with SmaCC.

Phase 2 - Verify and annotate AST (semantic analysis)

SemanticChecker in 'Compiler-Semantics' binds RBVariableNodes in the AST to temp, instance, or pool (class, pool, or global) SemVars, raising a notifier if no match is found.  The lookup is performed on a chain of SemScopes, one for the class, method and each nested closure.  Var usage (read/write/capture-in-closure) is tracked using a FiniteAutomaton.  The final usage state is used to determine unused temps and escaping temps (temps whose storage need to reside independent of the context since a block may assign to it after the context is gone).

Phase 3 - Translate AST to intermediate representation (IR)

SemanticTranslator in 'Compiler-Semantics' visits each node in the AST and invokes the appropriate instruction message on IRBuilder in 'Compiler-IR' resulting in an IRMethod.  An IRMethod consists of IRInstructions grouped by IRSequence (basic block).  Each IRInstruction represents a simple stack instruction like pushTemp: or send:.

Phase 4 - Optimize IR

Two simple optimization are applied to the IRMethod: absorbJumpsToSingleInstrs and absorbConstantConditionalJumps. The first eliminates jumps to returns and returns directly.  The second converts chained constant jumps created by and:/or: to single jumps directly to the target code.

Phase 5 - Translate IR to bytecodes (CompiledMethod)

IRTranslator in 'Compiler-IR' visits each IR instruction and invokes the appropriate bytecode message on BytecodeGenerator in 'Compiler-Bytecodes'.  Certain combinations of IR instructions (like storeTemp: and popTop) are mapped to a single bytecode instructions.
!


!Object methodsFor: 'evaluating' stamp: 'ajh 6/29/2004 14:40'!
doItScope

	^ self class parseScope instanceScope! !


!Behavior methodsFor: 'accessing' stamp: 'ajh 3/13/2003 21:13'!
decompilerClass
	"Answer a decompiler class appropriate for compiled methods of this class."

	^ Preferences compileBlocksAsClosures
		ifTrue: [Decompiler2]
		ifFalse: [Decompiler]! !

!Behavior methodsFor: 'compiling' stamp: 'ajh 6/23/2004 19:45'!
parseScope

	^ ClassScope new class: self! !

!Behavior methodsFor: 'private' stamp: 'ajh 5/14/2004 20:36'!
basicCompile: code notifying: requestor trailer: bytes ifFail: failBlock
	"Compile code without logging the source in the changes file"

	| methodNode |
	methodNode _ self compilerClass new
				compile: code
				in: self
				notifying: requestor
				ifFail: failBlock.
	Preferences compileBlocksAsClosures ifFalse: [
		methodNode encoder requestor: requestor].
	^ CompiledMethodWithNode generateMethodFromNode: methodNode trailer: bytes.! !


!ClassDescription methodsFor: 'compiling' stamp: 'ajh 3/11/2003 19:13'!
compile: code notifying: requestor trailer: bytes 
		ifFail: failBlock
		elseSetSelectorAndNode: selAndNodeBlock
	"Intercept this message in order to remember system changes.
	 5/15/96 sw: modified so that if the class does not wish its methods logged in the changes file, then they also won't be accumulated in the current change set.
	7/12/96 sw: use wantsChangeSetLogging to determine whether to put in change set"

	| methodNode selector newMethod priorMethodOrNil |
	methodNode _ self compilerClass new
				compile: code
				in: self
				notifying: requestor
				ifFail: failBlock.
	selector _ methodNode selector.
	selAndNodeBlock value: selector value: methodNode.
	requestor ifNotNil:
		["Note this change for recent submissions list"
		Utilities noteMethodSubmission: selector forClass: self].
	Preferences compileBlocksAsClosures ifFalse: [
		methodNode encoder requestor: requestor].  "Why was this not preserved?"
	newMethod _ methodNode generate: bytes.
	priorMethodOrNil _ (self methodDict includesKey: selector)
		ifTrue: [self compiledMethodAt: selector]
		ifFalse: [nil].
	Smalltalk changes noteNewMethod: newMethod forClass: self
		selector: selector priorMethod: priorMethodOrNil.
	self addSelector: selector withMethod: newMethod.
	^ newMethod! !


!CodeHolder methodsFor: 'message list' stamp: 'ajh 3/15/2003 16:03'!
decompiledSourceIntoContents
	"Obtain a source string by decompiling the method's code, and place that source string into my contents.  Get temps from source file"

	| tempNames class selector method |
	class _ self selectedClassOrMetaClass.
	selector _ self selectedMessageName.
	method _ class compiledMethodAt: selector ifAbsent: [^ ''].	"method deleted while in another project"
	currentCompiledMethod _ method.

	tempNames _ (class compilerClass new
					parse: method getSourceFromFile asString in: class notifying: nil)
					tempNames.
	contents _ ((method decompilerClass new withTempNames: tempNames)
			decompile: selector in: class method: method) decompileString.
	contents _ contents asText makeSelectorBoldIn: class.
	^ contents copy! !

!CodeHolder methodsFor: 'message list' stamp: 'ajh 3/15/2003 16:03'!
selectedMessage
	"Answer a copy of the source code for the selected message.  This generic version is probably actually never reached, since every subclass probably reimplements and does not send to super.  In time, ideally, most, or all, reimplementors would vanish and all would defer instead to a universal version right here.  Everything in good time."

	| class selector method tempNames |
	contents == nil ifFalse: [^ contents copy].
	self showingDecompile ifTrue:
		[^ self decompiledSourceIntoContents].

	class _ self selectedClassOrMetaClass.
	(class isNil or: [(selector _ self selectedMessageName) isNil]) ifTrue: [^ ''].
	method _ class compiledMethodAt: selector ifAbsent: [^ ''].	"method deleted while in another project"
	currentCompiledMethod _ method.

	(Sensor controlKeyPressed
		or: [method fileIndex > 0 and: [(SourceFiles at: method fileIndex) == nil]])
		ifTrue:
		["Emergency or no source file -- decompile without temp names"
		contents _ (method decompilerClass new decompile: selector in: class method: method)
			decompileString.
		contents _ contents asText makeSelectorBoldIn: class.
		^ contents copy].

	Sensor leftShiftDown ifTrue:
		["Special request to decompile -- get temps from source file"
		tempNames _ (class compilerClass new
						parse: method getSourceFromFile asString in: class notifying: nil)
						tempNames.
		contents _ ((method decompilerClass new withTempNames: tempNames)
				decompile: selector in: class method: method) decompileString.
		contents _ contents asText makeSelectorBoldIn: class.
		^ contents copy].

	self showComment
		ifFalse:
			[contents _ self sourceStringPrettifiedAndDiffed]
		ifTrue:
			[contents _ self commentContents].
	^ contents _ contents copy asText makeSelectorBoldIn: class! !


!Browser methodsFor: 'message list' stamp: 'ajh 3/15/2003 16:02'!
selectedMessage
	"Answer a copy of the source code for the selected message."

	| class selector method |
	contents == nil ifFalse: [^ contents copy].
	self showingDecompile ifTrue:
		[^ self decompiledSourceIntoContents].

	class _ self selectedClassOrMetaClass.
	selector _ self selectedMessageName.
	method _ class compiledMethodAt: selector ifAbsent: [^ ''].	"method deleted while in another project"
	currentCompiledMethod _ method.

	(Sensor controlKeyPressed
		or: [method fileIndex > 0 and: [(SourceFiles at: method fileIndex) == nil]])
		ifTrue:
		["Emergency or no source file -- decompile without temp names"
		contents _ (method decompilerClass new decompile: selector in: class method: method)
			decompileString.
		contents _ contents asText makeSelectorBoldIn: class.
		^ contents copy].

	Sensor leftShiftDown ifTrue:
		["Special request to decompile, old and dubious interface"
		^ self decompiledSourceIntoContents].

	self showingDocumentation
		ifFalse:
			[contents _ self sourceStringPrettifiedAndDiffed]
		ifTrue:
			[contents _ self commentContents].
	^ contents _ contents copy asText makeSelectorBoldIn: class! !


!CompiledMethod methodsFor: 'printing' stamp: 'ajh 3/15/2003 16:01'!
decompileString
	| clAndSel cl sel |
	clAndSel _ self who.
	clAndSel = #(unknown unknown)
		ifTrue:
			[cl _ Object.
			sel _ #xxxUnknown.
			self numArgs >= 1
				ifTrue:
					[sel _ sel , ':'.
					2 to: self numArgs do: [:i | sel _ sel , 'with:'].
					sel _ sel asSymbol]]
		ifFalse:
			[cl _ clAndSel first.
			sel _ clAndSel last].
	^ (self decompilerClass new
			decompile: sel in: cl method: self) decompileString! !

!CompiledMethod methodsFor: 'literals' stamp: 'ajh 5/20/2004 00:23'!
containsBlockClosures

	^ self embeddedBlockMethods size > 0! !

!CompiledMethod methodsFor: 'scanning' stamp: 'ajh 7/7/2004 18:20'!
readsField: varIndex 
	"Answer whether the receiver loads the instance variable indexed by the argument."

	self allEmbeddedBlockMethods do: [:meth |
		| scanner catcher instr scan |
		scanner _ meth scanner.
		catcher _ MessageCatcher new.
		[scanner atEnd] whileFalse: [
			instr _ scanner interpretNextInstructionFor: catcher.
			(instr selector = #send:super:numArgs: and: [instr argument = #privGetField:]) ifTrue: [
				scan _ scanner copy.
				scan pc: scan previousPc. scan pc: scan previousPc.
				instr _ scan interpretNextInstructionFor: catcher.
				(instr selector = #pushConstant: and: [instr argument = varIndex])
					ifTrue: [^ true]
			].
		].
	].
	self isReturnField ifTrue: [^ self returnField + 1 = varIndex].
	varIndex <= 16 ifTrue: [^ self scanFor: varIndex - 1].
	varIndex <= 64 ifTrue: [^ self scanLongLoad: varIndex - 1].
	^ self scanVeryLongLoad: 64 offset: varIndex - 1! !

!CompiledMethod methodsFor: 'scanning' stamp: 'ajh 5/30/2003 01:18'!
scanner

	^ InstructionStream on: self! !

!CompiledMethod methodsFor: 'scanning' stamp: 'ajh 7/7/2004 18:20'!
writesField: field 
	"Answer whether the receiver stores into the instance variable indexed by the argument."

	self allEmbeddedBlockMethods do: [:meth |
		| scanner catcher instr scan |
		scanner _ meth scanner.
		catcher _ MessageCatcher new.
		[scanner atEnd] whileFalse: [
			instr _ scanner interpretNextInstructionFor: catcher.
			(instr selector = #send:super:numArgs: and: [instr argument = #privStoreIn:field:]) ifTrue: [
				scan _ scanner copy.
				scan pc: scan previousPc. scan pc: scan previousPc.
				instr _ scan interpretNextInstructionFor: catcher.
				(instr selector = #pushConstant: and: [instr argument = field])
					ifTrue: [^ true]
			].
		].
	].
	self isQuick ifTrue: [^ false].
	field <= 8 ifTrue:
		[^ (self scanFor: 96 + field - 1) or: [self scanLongStore: field - 1]].
	field <= 64 ifTrue:
		[^ self scanLongStore: field - 1].
	^ self scanVeryLongStore: 160 offset: field - 1! !

!CompiledMethod methodsFor: 'source code management' stamp: 'ajh 3/15/2003 16:02'!
getSourceFor: selector in: class
	"Retrieve or reconstruct the source code for this method."
	| source flagByte |
	flagByte _ self last.
	(flagByte = 0
		or: [flagByte = 251 "some source-less methods have flag = 251, rest = 0"
			and: [((1 to: 3) collect: [:i | self at: self size - i]) = #(0 0 0)]])
		ifTrue:
		["No source pointer -- decompile without temp names"
		^ (self decompilerClass new decompile: selector in: class method: self)
			decompileString].
	flagByte < 252 ifTrue:
		["Magic sources -- decompile with temp names"
		^ ((self decompilerClass new withTempNames: self tempNames)
				decompile: selector in: class method: self)
			decompileString].

	"Situation normal;  read the sourceCode from the file"
	(source _ self getSourceFromFile) == nil ifFalse: [^ source].

	"Something really wrong -- decompile blind (no temps)"
	^ (self decompilerClass new decompile: selector in: class method: self)
			decompileString! !

!CompiledMethod methodsFor: 'decompiling' stamp: 'ajh 7/8/2004 19:33'!
ast
	"Return the block node for self"

	| homeMethodNode |
	homeMethodNode _ self homeMethodNode.
	self isBlockMethod ifFalse: [^ homeMethodNode].

	homeMethodNode ifNil: [^ self decompilerClass new decompileBlock: self].

	homeMethodNode compiledMethod.  "generate method"
	homeMethodNode nodesDo: [:node |
		(node isBlock and:
		 [node scope notNil "not-inlined" and:
		  [node compiledMethod = self]])
			ifTrue: [^ node]
	].

	"node not found, must have been compiled differently (compiler changed)"
	^ self decompilerClass new decompileBlock: self! !

!CompiledMethod methodsFor: 'decompiling' stamp: 'ajh 3/13/2003 21:13'!
decompilerClass

	^ self isClosureCompiled
		ifTrue: [Decompiler2]
		ifFalse: [Decompiler]! !

!CompiledMethod methodsFor: 'decompiling' stamp: 'ajh 6/28/2004 13:09'!
homeMethodNode
	"Return the parse tree of my home method"

	^ self homeMethodNodeDecompileClass: nil selector: nil! !

!CompiledMethod methodsFor: 'decompiling' stamp: 'ajh 7/7/2004 14:28'!
homeMethodNodeDecompileClass: class selector: selector
	"Return the parse tree of my home method"

	| source homeNode sel |
	(source _ self getSourceFromFile) ifNotNil: [
		sel _ selector.
		homeNode _ [self parserClass new
				parse: source readStream
				class: (class ifNil: [self sourceClass])
				noPattern: (sel = #DoIt)
				notifying: nil
				ifFail: [nil]
			] on: ParseError do: [:ex |
				sel ifNil: [sel _ #DoIt. ex retry]
					ifNotNil: [ex pass]]
	].
	^ homeNode ifNil: [
		self isBlockMethod
			ifTrue: [nil]
			ifFalse: [self decompileClass: class selector: selector]]! !

!CompiledMethod methodsFor: 'decompiling' stamp: 'ajh 3/2/2003 13:51'!
ir

	^ BytecodeDecompiler new decompile: self! !

!CompiledMethod methodsFor: 'decompiling' stamp: 'ajh 1/22/2003 16:49'!
parserClass

	^ self isClosureCompiled
		ifTrue: [Parser2]
		ifFalse: [Parser]! !


!CompiledMethod class methodsFor: 'instance creation' stamp: 'ajh 3/9/2003 15:09'!
primitive: primNum numArgs: numArgs numTemps: numTemps stackSize: stackSize literals: literals bytecodes: bytecodes trailer: trailerBytes
	"Create method with given attributes.  numTemps includes numArgs.  stackSize does not include numTemps."

	| compiledMethod |
	compiledMethod _ self
		newBytes: bytecodes size
		trailerBytes: trailerBytes 
		nArgs: numArgs
		nTemps: numTemps
		nStack: stackSize
		nLits: literals size
		primitive: primNum.
	(WriteStream with: compiledMethod)
		position: compiledMethod initialPC - 1;
		nextPutAll: bytecodes.
	literals withIndexDo: [:obj :i | compiledMethod literalAt: i put: obj].
	^ compiledMethod! !


!Compiler methodsFor: 'public access' stamp: 'ajh 7/8/2004 19:41'!
compileDoIt: textOrStream in: aContext to: receiver notifying: aRequestor ifFail: failBlock log: log
	"Compiles the sourceStream into a parse tree, then generates code into a method.  This method is then wrapped in a block with the receiver or context as the sole free variable which the method refers to.  If requestor is not nil, then it will receive a notify:at: message if there is a compile error, followed by the failBlock being executed."

	| scope parser blockNode method source |
	source _ textOrStream readStream.
	scope _ aContext
		ifNotNil: [aContext doItScope]
		ifNil: [receiver class parseScope instanceScope].
	parser _ Parser2 new.
	blockNode _ parser
			parse: source
			class: scope
			noPattern: true
			notifying: aRequestor
			ifFail: [^ failBlock value].
	method _ blockNode generate: #(0 0 0 0).

	log ifTrue: [
		method putSource: source contents
			fromParseNode: blockNode
			inFile: 2 
			withPreamble: [:file | file cr]].

	^ BlockClosure new
		env: (aContext ifNil: [receiver]);
		method: method;
		yourself! !

!Compiler methodsFor: 'public access' stamp: 'ajh 6/29/2004 14:31'!
evaluate2: textOrStream in: aContext to: receiver notifying: aRequestor ifFail: failBlock logged: logFlag

	| string |
	string _ textOrStream readStream upToEnd.
	^ (self compileDoIt: string in: aContext to: receiver notifying: aRequestor ifFail: [^ failBlock value] log: logFlag) value! !

!Compiler methodsFor: 'public access' stamp: 'ajh 7/8/2004 20:15'!
evaluate: textOrStream in: aContext to: receiver notifying: aRequestor ifFail: failBlock logged: logFlag
	"Compiles the sourceStream into a parse tree, then generates code into a 
	method. This method is then installed in the receiver's class so that it 
	can be invoked. In other words, if receiver is not nil, then the text can 
	refer to instance variables of that receiver (the Inspector uses this). If 
	aContext is not nil, the text can refer to temporaries in that context (the 
	Debugger uses this). If aRequestor is not nil, then it will receive a 
	notify:at: message before the attempt to evaluate is aborted. Finally, the 
	compiled method is invoked from here as DoIt or (in the case of 
	evaluation in aContext) DoItIn:. The method is subsequently removed 
	from the class, but this will not get done if the invocation causes an 
	error which is terminated. Such garbage can be removed by executing: 
	Smalltalk allBehaviorsDo: [:cl | cl removeSelector: #DoIt; removeSelector: 
	#DoItIn:]."

	| methodNode method value selector |
	(aContext
		ifNil: [Preferences compileBlocksAsClosures]
		ifNotNil: [aContext method isClosureCompiled]
	) ifTrue: [^ self evaluate2: textOrStream in: aContext to: receiver notifying: aRequestor ifFail: failBlock logged: logFlag].

	class _ (aContext == nil ifTrue: [receiver] ifFalse: [aContext receiver]) class.
	self from: textOrStream class: class context: aContext notifying: aRequestor.
	methodNode _ self translate: sourceStream noPattern: true ifFail:
		[^failBlock value].
	method _ methodNode generate: #(0 0 0 0).
	self interactive ifTrue:
		[method _ method copyWithTempNames: methodNode tempNames].
	
	selector _ context isNil
		ifTrue: [#DoIt]
		ifFalse: [#DoItIn:].
	class addSelectorSilently: selector withMethod: method.
	value _ context isNil
		ifTrue: [receiver DoIt]
		ifFalse: [receiver DoItIn: context].
	InMidstOfFileinNotification signal 
		ifFalse: [class basicRemoveSelector: selector].
	logFlag ifTrue: [SystemChangeNotifier uniqueInstance evaluated: sourceStream contents context: aContext].
	^ value! !

!Compiler methodsFor: 'private' stamp: 'ajh 3/6/2003 01:04'!
dialectParserClass

	^ self parserClass = Parser2
		ifTrue: [self toDo]
		ifFalse: [DialectParser]! !

!Compiler methodsFor: 'private' stamp: 'ajh 7/8/2004 17:47'!
parserClass: aParserClass

	parserClass _ aParserClass.
! !


!Compiler class methodsFor: 'accessing' stamp: 'ajh 1/20/2002 18:57'!
parserClass
	"Return a parser class to use for parsing method headers."

	^ Preferences compileBlocksAsClosures
		ifTrue: [Parser2]
		ifFalse: [Parser]! !

!Compiler class methodsFor: 'as yet unclassified' stamp: 'ajh 3/26/2003 00:39'!
initialize

	Preferences
		addPreference: #compileBlocksAsClosures
		category: #general
		default: false
		balloonHelp: 'If true, blocks ([...]) will be compiled as BlockClosures instead of BlockContexts.  BlockClosures are independent of their home context but are a little slower (for now).  Only newly compiled methods will be affected by this preference.'
! !


!ContextPart methodsFor: 'accessing' stamp: 'ajh 6/28/2004 14:56'!
methodNode

	| selector methodClass |
	self method isBlockMethod ifTrue: [^ self method ast].
	selector _ self receiver class
		selectorAtMethod: self method
		setClass: [:mclass | methodClass _ mclass].
	^ self method homeMethodNodeDecompileClass: methodClass selector: selector! !


!Debugger methodsFor: 'private' stamp: 'ajh 7/7/2004 15:48'!
contextStackIndex: anInteger oldContextWas: oldContext 
	"Change the context stack index to anInteger, perhaps in response to user selection."

	| newMethod c |
	contextStackIndex := anInteger.
	anInteger = 0
		ifTrue: [currentCompiledMethod := theMethodNode := tempNames := sourceMap := contents := nil.
			self changed: #contextStackIndex.
			self decorateButtons.
			self contentsChanged.
			contextVariablesInspector object: nil.
			self receiverInspectorObject: self receiver context: nil.
			^ self].
	(newMethod := oldContext == nil
					or: [oldContext method ~~ (currentCompiledMethod := self selectedContext method)])
		ifTrue: [tempNames := sourceMap := nil.
			theMethodNode := Preferences browseWithPrettyPrint
				ifTrue: [ 	self selectedContext methodNodeFormattedAndDecorated: Preferences colorWhenPrettyPrinting ]
				ifFalse: [	self selectedContext methodNode ].
			contents := self selectedMessage.
			self contentsChanged.
			self pcRange
			"will compute tempNamesunless noFrills"].
	self changed: #contextStackIndex.
	self decorateButtons.
	tempNames == nil
		ifTrue: [tempNames := self selectedClassOrMetaClass parserClass new parseArgsAndTemps: contents notifying: nil].
	contextVariablesInspector object: (c _ self selectedContext).
	self receiverInspectorObject: self receiver context: c.
	newMethod
		ifFalse: [self changed: #contentsSelection]! !

!Debugger methodsFor: 'private' stamp: 'ajh 7/7/2004 15:56'!
receiverInspectorObject: obj context: ctxt

	"set context before object so it can refer to context when building field list"
	receiverInspector context: ctxt.
	receiverInspector object: obj.
! !


!MethodContext methodsFor: 'closure support' stamp: 'ajh 7/8/2004 17:43'!
doItScope
	"scope (environment) for expressions executed within a method context. self will be the receiver of the do-it method. We want temp vars directly accessible"

	^ self methodNode scope asDoItScope! !


!ParseNode methodsFor: 'testing' stamp: 'ajh 6/29/2004 14:13'!
isDoIt
	"polymorphic with RBNodes; called by debugger"

	^ false! !

!ParseNode methodsFor: 'testing' stamp: 'ajh 3/24/2003 16:16'!
isJust: node

	^ false! !

!ParseNode methodsFor: 'converting' stamp: 'ajh 3/24/2003 16:25'!
ifNilReceiver

	^ self! !


!MethodNode methodsFor: 'code generation' stamp: 'ajh 3/24/2003 14:55'!
generate: trailer 

	Preferences compileBlocksAsClosures ifTrue: [
		^ (Parser2 new
			parse: self decompileString
			class: self encoder classEncoding
		   ) generate: trailer
	].

	^ self generateNative: trailer! !

!MethodNode methodsFor: 'code generation' stamp: 'ajh 3/24/2003 14:52'!
sourceMap
	"Answer a SortedCollection of associations of the form: pc (byte offset in 
	me) -> sourceRange (an Interval) in source text."

	self generateNative: #(0 0 0 0).
	^encoder sourceMap! !


!Parser methodsFor: 'public access' stamp: 'ajh 6/29/2004 14:31'!
parse: sourceStream class: class noPattern: noPattern notifying: req ifFail: aBlock

	^ self parse: sourceStream class: class noPattern: noPattern context: nil notifying: req ifFail: aBlock! !


!Scanner class methodsFor: 'testing' stamp: 'ajh 3/24/2003 14:10'!
isLiteralSymbol: aSymbol 
	"Test whether a symbol can be stored as # followed by its characters.  
	Symbols created internally with asSymbol may not have this property, 
	e.g. '3' asSymbol."
	| i ascii type |
	i _ aSymbol size.
	i = 0 ifTrue: [^ false].
	ascii _ (aSymbol at: 1) asciiValue.
	"TypeTable should have been origined at 0 rather than 1 ..."
	ascii = 0 ifTrue: [^ false].
	type _ TypeTable at: ascii.
	(type == #xBinary or: [type == #verticalBar]) ifTrue: 
			[[i > 1]
				whileTrue: 
					[ascii _ (aSymbol at: i) asciiValue.
					ascii = 0 ifTrue: [^ false].
					type _ TypeTable at: ascii.
					(type == #xBinary or: [type == #verticalBar]) ifFalse: [^ false].
					i _ i - 1].
			^ true].
	type == #xLetter ifTrue: 
			[[i > 1]
				whileTrue: 
					[ascii _ (aSymbol at: i) asciiValue.
					ascii = 0 ifTrue: [^ false].
					type _ TypeTable at: ascii.
					(type == #xLetter or: [type == #xDigit or: [type == #xColon]])
						ifFalse: [^ false].
					i _ i - 1].
			^ true].
	^ false! !


!String methodsFor: 'private' stamp: 'ajh 3/18/2003 13:16'!
correctAgainstEnumerator: wordBlock continuedFrom: oldCollection
	"The guts of correction, instead of a wordList, there is a block that should take another block and enumerate over some list with it."

	| choices scoreMin results score maxChoices |
	scoreMin _ self size // 2 min: 3.
	maxChoices _ 10.
	oldCollection isNil
		ifTrue: [ choices _ SortedCollection sortBlock: [ :x :y | x value > y value ] ]
		ifFalse: [ choices _ oldCollection ].
	wordBlock isNil
		ifTrue:
			[ results _ OrderedCollection new.
			1 to: (maxChoices min: choices size) do: [ :i | results add: (choices at: i) key ] ]
		ifFalse:
			[ wordBlock value: [ :word |
				((score _ self alike: word) >= scoreMin
				 and: [(choices anySatisfy: [:assoc | assoc key = word]) not])
				ifTrue: [ choices add: (Association key: word value: score).
					     (choices size >= maxChoices) ifTrue: [ scoreMin _ (choices at: maxChoices) value] ] ].
			results _ choices ].
	^ results! !


!Text methodsFor: 'emphasis' stamp: 'ajh 7/6/2004 23:25'!
makeSelectorBold
	"For formatting Smalltalk source code, set the emphasis of that portion of 
	the receiver's string that parses as a message selector to be bold."

	| parser i |
	string size = 0 ifTrue: [^ self].
	i _ 0.
	[(string at: (i _ i + 1)) isSeparator] whileTrue.
	(string at: i) = $[ ifTrue: [^ self].  "block, no selector"
	[(parser _ Compiler parserClass new) parseSelector: string] on: Error do: [^ self].
	self makeBoldFrom: 1 to: (parser endOfLastToken min: string size)! !


!UndefinedObject methodsFor: 'class hierarchy' stamp: 'ajh 6/23/2004 20:40'!
parseScope

	^ GlobalScope new! !

!UndefinedObject methodsFor: 'class hierarchy' stamp: 'ajh 1/21/2003 13:03'!
parserClass

	^ Compiler parserClass! !

MethodContext removeSelector: #methodNode!
Message class removeSelector: #catcher!
Compiler initialize!
Compiler removeSelector: #cacheDoItNode:!
CompiledMethod removeSelector: #blockNode!
CompiledMethod removeSelector: #blockNodeIn:!
CompiledMethod removeSelector: #methodNode!
CompiledMethod removeSelector: #methodNodeDecompileClass:selector:!
